from factory_sdk.dto.model import ModelMeta, ModelRevision, ModelObject
from factory_sdk.dto.dataset import DatasetMeta, DatasetRevision, DatasetObject
from factory_sdk.dto.metric import MetricMeta, MetricRevision, MetricObject
from factory_sdk.dto.preprocessor import (
    PreprocessorMeta,
    PreprocessorRevision,
    PreprocessorObject,
)
from factory_sdk.dto.resource import FactoryRevisionRef
from typing import List
from factory_sdk.dto.task import TrainingTask
from factory_sdk.dto.adapter import AutoTrainingParams
from factory_sdk.fast.inspect import get_cleaned_module_source, hash_code_by_ast
from factory_sdk.exceptions.api import NotFoundException
from factory_sdk.dto.adapter import AdapterMeta, AdapterInitData, AdapterRevision
from rich import print
from typing import Optional
from factory_sdk.dto.resource import FactoryMetaState, FactoryRevisionState
from hashlib import md5
from factory_sdk.dto.recipe import RecipeObject
from factory_sdk.datasets import download_dataset
from factory_sdk.dto.recipe import RecipeRevision
from factory_sdk.dto.adapter import AdapterArgs, TrainArgs, InitArgs
from factory_sdk.recipe import download_recipe
from factory_sdk.models import download_model
from factory_sdk.logging import print_exceptions


class Adapters:
    def __init__(self, client):
        self.client = client
        self._name = None
        self._recipe: RecipeObject = None
        self._adapter_args: AdapterArgs
        self._train_args: TrainArgs
        self._init_args: InitArgs

    def with_name(self, name: str):
        self._name = name
        return self

    def based_on_recipe(self, recipe: RecipeObject):
        self._recipe = recipe
        return self

    def using_model(self, model: ModelObject):
        self._model = model
        return self

    def with_hyperparameters(
        self,
        train_args: TrainArgs = TrainArgs(),
        adapter_args: AdapterArgs = AdapterArgs(),
        InitArgs: InitArgs = InitArgs(),
    ):
        self._adapter_args = adapter_args
        self._train_args = train_args
        self._init_args = InitArgs
        return self

    @print_exceptions(show_locals=False)
    def train(self):
        recipe_revision: RecipeRevision = self._recipe.get_revision()

        ##
        dataset_ref = None
        for dep in recipe_revision.dependencies:
            if dep.type == "dataset":
                dataset_ref = dep
                break
        assert dataset_ref is not None, "Dataset not found in recipe dependencies"

        #### Download the dataset ####

        model_path = download_model(
            self._model.meta.id, self._model.revision, self.client, return_path=True
        )

        dataset_path = download_dataset(
            dataset_ref.id, dataset_ref.revision, self.client, return_path=True
        )

        recipe_path = download_recipe(
            self._recipe.meta.id, recipe_revision.id, self.client, return_path=True
        )

        from factory_sdk.fast.train.initialize import initialize_adapter
        from factory_sdk.train.start import start_training

        run_id, run_path = initialize_adapter(
            self._adapter_args,
            self._train_args,
            self._init_args,
            model_path,
            dataset_path,
            recipe_path,
        )
        client_params = self.client.get_init_params()
        start_training(
            model_path,
            dataset_path,
            recipe_path,
            self._recipe.meta.id,
            recipe_revision.id,
            run_path,
            client_params,
            self._name,
        )
