from factory_sdk.fast.inspect import (
    get_cleaned_module_source,
    hash_code_by_ast,
    load_code_from_string,
    find_dependencies,
)
import json
from hashlib import md5
from factory_sdk.dto.metric import (
    MetricMeta,
    MetricRevision,
    MetricInitData,
    MetricCallObject,
    MetricObject,
)
from factory_sdk.exceptions.api import NotFoundException
from typing import Optional
from factory_sdk.dto.resource import FactoryMetaState, FactoryRevisionState
from rich import print
from tempfile import TemporaryDirectory
from glob import glob
from joblib import Parallel, delayed
import os
from factory_sdk.dto.model import Token, GeneratedToken, MetricScore
from tqdm.auto import tqdm


class Metrics:
    def __init__(self, client):
        self.client = client

    def with_name(self, name):
        return MetricAdapter(self.client, name)


class MetricAdapter:
    def __init__(self, client, name):
        self.client = client
        self.name = name

    def test_implementation(self, fn):
        def gen_test_case(length):
            label = [Token(id=i) for i in range(length)]
            generated = [
                GeneratedToken(id=i, logprob=-1.0, rank=1) for i in range(length)
            ]
            return label, generated

        pbar = tqdm(total=184, desc="Running tests", leave=False)

        # Test various sequence lengths (0 to 5001)
        for length in range(0, 5001, 42):
            label, generated = gen_test_case(length)
            result = fn(label, generated)
            assert isinstance(result, MetricScore)
            pbar.update(1)

        # Test different length combinations
        lens = [0, 1, 20, 50, 100, 2000, 5000]
        for len1 in lens:
            for len2 in lens:
                label = [Token(id=i) for i in range(len1)]
                generated = [
                    GeneratedToken(id=i, logprob=-1.0, rank=1) for i in range(len2)
                ]
                result = fn(label, generated)
                assert isinstance(result, MetricScore)
                pbar.update(1)

        # Test with different token values but same lengths
        for length in [1, 20, 500, 1000]:
            # Same length, different idx
            label = [Token(id=i) for i in range(length)]
            generated = [
                GeneratedToken(id=i + 1, logprob=-1.0, rank=1) for i in range(length)
            ]
            result = fn(label, generated)
            assert isinstance(result, MetricScore)
            pbar.update(1)

            # Different logprob values
            generated = [
                GeneratedToken(id=i, logprob=float(-i), rank=1) for i in range(length)
            ]
            result = fn(label, generated)
            assert isinstance(result, MetricScore)
            pbar.update(1)

            # Different rank values
            generated = [
                GeneratedToken(id=i, logprob=-1.0, rank=i + 1) for i in range(length)
            ]
            result = fn(label, generated)
            assert isinstance(result, MetricScore)
            pbar.update(1)

        # Edge cases
        # All same values
        length = 1000
        label = [Token(id=0)] * length
        generated = [GeneratedToken(id=0, logprob=-1.0, rank=1)] * length
        result = fn(label, generated)
        assert isinstance(result, MetricScore)
        pbar.update(1)

        # Alternating values
        label = [Token(id=i % 2) for i in range(length)]
        generated = [
            GeneratedToken(id=i % 2, logprob=-1.0, rank=1) for i in range(length)
        ]
        result = fn(label, generated)
        assert isinstance(result, MetricScore)
        pbar.update(1)

        # Reversed sequences
        label = [Token(id=i) for i in range(length)]
        generated = [
            GeneratedToken(id=length - i - 1, logprob=-1.0, rank=1)
            for i in range(length)
        ]
        result = fn(label, generated)
        assert isinstance(result, MetricScore)
        pbar.update(1)

        pbar.close()

    def from_code(self, ref):
        # Extract code and compute fingerprint
        path, fn_name, code__src = get_cleaned_module_source(ref)

        fingerprint = hash_code_by_ast(code__src)

        ##get metric from factory
        try:
            metric: MetricMeta = self.client.get(
                f"tenants/{self.client._tenant.id}/projects/{self.client._project.id}/metrics/{self.name}",
                response_class=MetricMeta,
                scope="names",
            )
            if metric.last_revision is not None:
                revision: MetricRevision = self.client.get(
                    f"metrics/{metric.id}/revisions/{metric.last_revision}",
                    response_class=MetricRevision,
                )

                if revision.fingerprints.get("code") == fingerprint:
                    return MetricObject(meta=metric, revision=revision)

        except NotFoundException:
            metric = None

        # load the code from code_src (string)
        reloaded_fn = load_code_from_string(code__src, fn_name)

        # Create a temporary directory to store code and params
        with TemporaryDirectory() as dir:
            # test the implementation
            print("[green]🧪 Testing the preprocessor implementation...[/green]")
            self.test_implementation(reloaded_fn)
            print("[bold green]🎉 Metric implementation passed all tests![/bold green]")

            # Write the source code
            code_file = f"{dir}/code.py"
            with open(code_file, "w") as f:
                f.write(code__src)

            packages = find_dependencies(code_file)

            print(packages)
            exit()

        return MetricObject(meta=metric, revision=revision)
