import os
import subprocess
import json
import sys
import pty
import select
import fcntl
import termios

def start_training(
    model_path,
    dataset_path,
    recipe_path,
    recipe_id,
    recipe_revision,
    run_path,
    client_params,
    adapter_name,
):
    # locate path of run.py
    run_file_path = os.path.join(os.path.dirname(__file__), "run.py")
    
    # Create PTY
    master_fd, slave_fd = pty.openpty()
    
    # Set PTY to raw mode to prevent input processing
    mode = termios.tcgetattr(master_fd)
    mode[3] = mode[3] & ~termios.ECHO  # Disable echo
    termios.tcsetattr(master_fd, termios.TCSAFLUSH, mode)
    
    # Make master non-blocking
    flags = fcntl.fcntl(master_fd, fcntl.F_GETFL)
    fcntl.fcntl(master_fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)

    # Start deepspeed process
    process = subprocess.Popen(
        [
            "deepspeed",
            run_file_path,
            "--model_path", model_path,
            "--dataset_path", dataset_path,
            "--recipe_path", recipe_path,
            "--run_path", run_path,
            "--client_params", json.dumps(client_params),
            "--adapter_name", adapter_name,
            "--recipe_id", recipe_id,
            "--recipe_revision", recipe_revision,
        ],
        stdout=slave_fd,
        stderr=slave_fd,
        env={
            **os.environ,
            "PYTHONUNBUFFERED": "1",
            "TERM": os.environ.get("TERM", "xterm-256color"),  # Preserve colors
        }
    )
    
    # Close slave fd as it's now owned by the child process
    os.close(slave_fd)
    
    try:
        # Read output and forward it to stdout
        while True:
            try:
                # Use select to wait for data
                rlist, _, _ = select.select([master_fd], [], [], 0.1)
                
                if master_fd in rlist:
                    data = os.read(master_fd, 1024)
                    if not data:  # EOF
                        break
                    sys.stdout.buffer.write(data)
                    sys.stdout.buffer.flush()
                
                # Check if process has finished
                if process.poll() is not None:
                    break
                    
            except (OSError, IOError) as e:
                if e.errno != errno.EAGAIN:  # Ignore would-block errors
                    raise
                
    finally:
        # Clean up
        os.close(master_fd)
        
        # Make sure process is terminated
        if process.poll() is None:
            process.terminate()
            try:
                process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                process.kill()
                process.wait()
    
    return process.returncode